#include <stdio.h>
#include <stdlib.h>
#include "rsaeuro.h"
#include "mqsec.h"

void GetDigest (unsigned char *digestOut, unsigned int *digestLen, unsigned char *buf, unsigned int bufLen);
int CreateKeyPair (R_RSA_PUBLIC_KEY * pubk, R_RSA_PRIVATE_KEY * privk, unsigned char *buf, unsigned int buflen);
// unsigned char buffer[4096];
MQ_QMGR_DATA mqQmgrData[80];
unsigned int noOfQmgrs;
MQ_CHANNEL_DATA mqChannelData[1024];
unsigned int noOfChannels;
R_RSA_PUBLIC_KEY *cssdPubk;
R_RSA_PRIVATE_KEY *cssdPrivk;
static void InitQmgrFromFile (void);
static void InitChannelsFromFile (void);
static void SetKeys (MQ_QMGR_DATA * qmgrsec);
static void DisplayQmgrs (MQ_QMGR_DATA * qmgrsec);
static void DisplayChannels (MQ_CHANNEL_DATA * channeldata);
static void CreateFiles (MQ_QMGR_DATA * qmgrsec);
static void WritePrivateFile (unsigned char *fname, MQ_QMGR_DATA * qmgrsec);
static void WritePublicFile (unsigned char *fname, MQ_QMGR_DATA * qmgrsec);
static unsigned char *GetRecord (unsigned char *buf, unsigned int maxlen, FILE * fptptr);
static MQ_QMGR_DATA *GetQmgrData (unsigned char *qmgr);
static void PrintBuff (unsigned char *out, int len);

int
main (int argc, char **argv, char **envp)
{
  InitQmgrFromFile ();
  InitChannelsFromFile ();
  CreateFiles (mqQmgrData);
  return 0;
}

static void
InitQmgrFromFile (void)
{
  FILE *fPtr;
  MQ_QMGR_DATA *qmgrPtr = mqQmgrData + 1;
/* Initialize with 1 because we do CSSD outside the loop */
  unsigned int i = 1;
  unsigned char buf[256];
  unsigned char *bufPtr;
  fPtr = fopen ("QMGRS.INP", "r");
  if (fPtr == NULL) {
    printf ("Cannot open QMGRS.INP\n");
    exit (8);
  }
  strcpy (mqQmgrData[0].qmgr, "CSSD");
  SetKeys (mqQmgrData);
  while (GetRecord (buf, 256, fPtr)) {
    strcpy (qmgrPtr->qmgr, buf);
    SetKeys (qmgrPtr);
//PrintBuff( (unsigned char*) qmgrPtr->qmgr, MQ_Q_MGR_NAME_LENGTH );
    qmgrPtr += 1;		/* point to next element in array */
/* Make sure we won't overrun the array */
    if (i++ == 80)
      break;
  }
  noOfQmgrs = i;
  fclose (fPtr);
  cssdPubk = &mqQmgrData[0].pubk;
  cssdPrivk = &mqQmgrData[0].privk;
/* Only for debugging */
  DisplayQmgrs (mqQmgrData);
}

static void
InitChannelsFromFile (void)
{
  unsigned char buf[256];
  unsigned char *bufPtr;
  FILE *fPtr;
  MQ_CHANNEL_DATA *channelPtr = mqChannelData;
  unsigned int i = 0;
  unsigned char *token;
// unsigned char seps[] = " ";
  fPtr = fopen ("CHANNELS.INP", "r");
  if (fPtr == NULL) {
    perror ("Cannot open CHANNELS.INP\n");
    exit (8);
  }
  while (GetRecord (buf, 256, fPtr)) {
    strcpy (channelPtr->channel, buf);
    GetRecord (buf, 256, fPtr);
    strcpy (channelPtr->qmgrA, buf);
    GetRecord (buf, 256, fPtr);
    strcpy (channelPtr->qmgrB, buf);
    printf ("Channel %s, qmgrA %s, qmgrB %s\n", channelPtr->channel, channelPtr->qmgrA, channelPtr->qmgrB);
    channelPtr += 1;		/* point to next element in array */
/* Make sure we won't overrun the array */
    if (i++ == 1024)
      break;
  }
  noOfChannels = i;
  fclose (fPtr);
/* Only for debugging */
  DisplayChannels (mqChannelData);
}

static void
SetKeys (MQ_QMGR_DATA * qmgrsec)
{
  unsigned int status;
/* Initialize the struct */
  qmgrsec->qmgrLen = strlen (qmgrsec->qmgr) + 1;
/* Get rid of the newline character in the string */
  qmgrsec->qmgrLen -= 1;
  qmgrsec->qmgr[qmgrsec->qmgrLen] = 0x00;
  printf ("Qmgrname: %s Length: %d\n", qmgrsec->qmgr, qmgrsec->qmgrLen);
  status = CreateKeyPair (&(qmgrsec->pubk), &(qmgrsec->privk), qmgrsec->qmgr, qmgrsec->qmgrLen);
  if (status)
    printf ("Key generation failed with return code: %02lX\n", status);
}

static void
DisplayQmgrs (MQ_QMGR_DATA * qmgrsec)
{
  printf ("%d queue manager names read from file\n", noOfQmgrs);
  while (qmgrsec->qmgr[0] != 0) {
    printf ("%-14s %d \n", qmgrsec->qmgr, qmgrsec->qmgrLen);
    qmgrsec += 1;
  }
}

static void
DisplayChannels (MQ_CHANNEL_DATA * channeldata)
{
  printf ("%d channel names read from file\n", noOfChannels);
  while (channeldata->channel[0] != 0) {
    printf ("%-20s %14s %14s \n", channeldata->channel, channeldata->qmgrA, channeldata->qmgrB);
    channeldata += 1;
  }
}

int
CreateKeyPair (R_RSA_PUBLIC_KEY * pubk, R_RSA_PRIVATE_KEY * privk, unsigned char *buf, unsigned int buflen)
{
  R_RANDOM_STRUCT random;
  R_RSA_PROTO_KEY protoKey;
  int status;
  unsigned char digestOut[MAX_DIGEST_LEN];
  unsigned int digestLen;
  R_RandomInit (&random);
/* Key size in bits */
  protoKey.bits = 512;
  protoKey.useFermat4 = 1;
  GetDigest (digestOut, &digestLen, buf, buflen);
  status = R_RandomUpdate (&random, digestOut, digestLen);
  R_RandomFinal (&random);
//R_RandomCreate(&random);
  status = R_GeneratePEMKeys (pubk, privk, &protoKey, &random);
  printf ("R_GeneratePEMKeys() ==> %d\n", status);
//PrintBuff( (unsigned char*) pubk, sizeof(R_RSA_PUBLIC_KEY) );
//PrintBuff( (unsigned char*) privk, sizeof(R_RSA_PRIVATE_KEY) );
  return status;
}

/*
* Generating a message digest of memory-resident data
*/
void
GetDigest (unsigned char *digestOut, unsigned int *digestLen, unsigned char *buf, unsigned int bufLen)
{
  R_DIGEST_CTX md5ctxt;
  unsigned int i;
  R_DigestBlock (digestOut, digestLen, buf, bufLen, DA_MD5);
}

static void
CreateFiles (MQ_QMGR_DATA * qmgrsec)
{
  unsigned char privFileName[256];
  unsigned char publicFileName[256];
  unsigned int i;
  for (i = 1; i < noOfQmgrs; i++) {
/* Build file private file name */
    strcpy (privFileName, qmgrsec[i].qmgr);
    strcat (privFileName, ".PRV");
    WritePrivateFile (privFileName, &(qmgrsec[i]));
    strcpy (publicFileName, qmgrsec[i].qmgr);
    strcat (publicFileName, ".PUB");
    WritePublicFile (publicFileName, &(qmgrsec[i]));
  }
}

static void
WritePrivateFile (unsigned char *fname, MQ_QMGR_DATA * qmgrsec)
{
  FILE *fptr;
  unsigned int len;
  unsigned char digestOut[MAX_DIGEST_LEN];
  unsigned int digestLen;
  printf ("Writing private file %s\n", fname);
  if ((fptr = fopen (fname, "wb")) == NULL) {
    printf ("Cannot open file %s\n", fname);
    exit (16);
  }
/* We always prepend data with an appropiate eye catcher */
/* Write queue manager's name to file */
  fwrite (QMGR_NAME_EYE_CATCHER, QMGR_NAME_EYE_CATCHER_SIZE, 1, fptr);
  fwrite (qmgrsec->qmgr, MQ_Q_MGR_NAME_LENGTH, 1, fptr);
/* Write the CSSD's public key to file */
  fwrite (CSSD_PUBK_EYE_CATCHER, CSSD_PUBK_EYE_CATCHER_SIZE, 1, fptr);
  fwrite (&(mqQmgrData[0].pubk), sizeof (R_RSA_PUBLIC_KEY), 1, fptr);
/* Write the queue manager's public key to file */
  fwrite (QMGR_PUBK_EYE_CATCHER, QMGR_PUBK_EYE_CATCHER_SIZE, 1, fptr);
  fwrite (&(qmgrsec->pubk), sizeof (R_RSA_PUBLIC_KEY), 1, fptr);
/* Write the queue manager's private key to file */
  fwrite (QMGR_PRIVK_EYE_CATCHER, QMGR_PRIVK_EYE_CATCHER_SIZE, 1, fptr);
  fwrite (&(qmgrsec->privk), sizeof (R_RSA_PRIVATE_KEY), 1, fptr);
  fclose (fptr);
}

static void
WritePublicFile (unsigned char *fname, MQ_QMGR_DATA * qmgrsec)
{
  FILE *fptr;
  unsigned int len;
  unsigned char signature[MAX_SIGNATURE_LEN];
  unsigned int signatureLen;
  unsigned int i;
  unsigned char *qmgrPtr;
  MQ_QMGR_DATA *qmgrdata;
  R_RSA_PUBLIC_KEY *pubk;
  unsigned char buf[1024];
  int status;
  R_SIGNATURE_CTX context;
  printf ("Writing public file %s\n", fname);
  if ((fptr = fopen (fname, "wb")) == NULL) {
    printf ("Cannot open file %s\n", fname);
    exit (16);
  }
  for (i = 0; i < noOfChannels; i++) {
    printf ("A: %s B: %s Q: %s\n", mqChannelData[i].qmgrA, mqChannelData[i].qmgrB, qmgrsec->qmgr);
    if (strcmp (mqChannelData[i].qmgrA, qmgrsec->qmgr) == 0) {
      qmgrdata = GetQmgrData (mqChannelData[i].qmgrB);
      if (qmgrdata == NULL) {
	printf ("Couldn't find public key of qmgr %s\n", mqChannelData[i].qmgrB);
	return;
      }
    }
    else if (strcmp (mqChannelData[i].qmgrB, qmgrsec->qmgr) == 0) {
      qmgrdata = GetQmgrData (mqChannelData[i].qmgrA);
      if (qmgrdata == NULL) {
	printf ("Couldn't find public key of qmgr %s\n", mqChannelData[i].qmgrA);
	return;
      }
    }

    else
      continue;;
/* Write channel name */
    fwrite (CHANNEL_NAME_EYE_CATCHER, CHANNEL_NAME_EYE_CATCHER_SIZE, 1, fptr);
    fwrite (mqChannelData[i].channel, MQ_CHANNEL_NAME_LENGTH, 1, fptr);
/* Write partner qmgr name */
    fwrite (QMGR_NAME_EYE_CATCHER, QMGR_NAME_EYE_CATCHER_SIZE, 1, fptr);
    fwrite (qmgrdata->qmgr, MQ_Q_MGR_NAME_LENGTH, 1, fptr);
/* Write partner qmgr's public key */
    fwrite (QMGR_PUBK_EYE_CATCHER, QMGR_PUBK_EYE_CATCHER_SIZE, 1, fptr);
    fwrite (&qmgrdata->pubk, sizeof (R_RSA_PUBLIC_KEY), 1, fptr);
/* Create and write a digital signature */
    status = R_SignInit (&context, DA_MD5);
    status = R_SignUpdate (&context, mqChannelData[i].channel, MQ_CHANNEL_NAME_LENGTH);
    status = R_SignUpdate (&context, qmgrdata->qmgr, MQ_Q_MGR_NAME_LENGTH);
    status = R_SignUpdate (&context, (unsigned char *)
			   &qmgrdata->pubk, sizeof (R_RSA_PUBLIC_KEY));
    status = R_SignFinal (&context, signature, &signatureLen, cssdPrivk);
    fwrite (DIGSIG_EYE_CATCHER, DIGSIG_EYE_CATCHER_SIZE, 1, fptr);
    fwrite (&signatureLen, sizeof (unsigned int), 1, fptr);
    fwrite (signature, MAX_SIGNATURE_LEN, 1, fptr);
  }
  fclose (fptr);
}

/* Returns a pointer to the QMGR_DATA structure */
MQ_QMGR_DATA *
GetQmgrData (unsigned char *qmgr)
{
  unsigned int i;
/* We begin with i=1 because we skip the CSSD */
  for (i = 1; i < noOfQmgrs; i++) {
    if (strcmp (qmgr, mqQmgrData[i].qmgr) == 0)
      return mqQmgrData + i;
  }
  return 0;
}

static unsigned char *
GetRecord (unsigned char *buf, unsigned int maxlen, FILE * fptr)
{
  unsigned char *bufPtr;
  while (1) {
    if ((bufPtr = fgets (buf, maxlen, fptr)) == 0)
      return bufPtr;
    if (buf[0] == ' ' || buf[0] == '*' || buf[0] == '\n')
      continue;
/* Remove trailing \n */
    buf[strlen (buf) - 1] = 0x00;
/* Removing trailing blanks */
    while (*buf != ' ') {
      if (*buf == 0x00)
	break;
      buf += 1;
    }
    *buf = 0x00;
    return bufPtr;
  }
}

static void
PrintBuff (unsigned char *out, int len)
{
  int i;
  for (i = 0; i < len; i++)
    printf ("%02X", *(out + i));
  printf ("\n");
}
